#!/usr/bin/ruby -w
#
# DNS Key
#
# By:  ͵ <yokota@netlab.is.tsukuba.ac.jp>
#
# $Id: dnskey.rb,v 1.1 2002/10/15 07:19:35 elca Exp elca $

if ENV["ROOT"] != nil
  $LOAD_PATH.unshift(ENV["ROOT"])
end

require 'socket'
require 'thread'
require 'getopts'

require "datatype.rb"
require "util.rb"

require "multilog.rb"
require "cache.rb"
require "dns_parser"

require "md5_auth.rb"
require "pgp_auth.rb"

ML = MultiLog.new
ML.open()

(ML.log("UPLINK env");   exit 111) if ENV["UPLINK"] == nil
(ML.log("DOWNLINK env"); exit 111) if ENV["DOWNLINK"] == nil

UPLINK   = ENV["UPLINK"]
DOWNLINK = ENV["DOWNLINK"]

ML.log("DNS downlink: " + DOWNLINK)
ML.log("DNS uplink:   " + UPLINK)

##########################################################################

def make_reject_packet(packets)
  query = packets["query"][0]
  header = packets["packet"]

  dom = "\xc0\x0c"
  txt = str_to_dnstxt("auth failed")
  ttl = "\0\0\0\5"
  len = [txt.length].pack("n")

  r = header[0,4] + "\0\1" + "\0\0" + "\0\1" + "\0\0" + \
      query["q"] + query["q_type"] + query["q_class"] + \
      dom + DnsType::TXT + DnsClass::INET + ttl + len + txt

  r[2] |= 0x84  # answer & authenticated
  r[3] &= ~0x0f
  r[3] |= 0x03  # no query error
  #r[3] |= 0x09  # not auth error

  return r
end

def make_key(p)
  return p["id"]+p["query"][0]["q"]
end

def num_records_str(p)
  return (p["num_q"] +
	  p["ans_rr"] +
	  p["ort_rr"] +
	  p["add_rr"]).unpack("nnnn").join("+")
end

def header_str(p)
  type = p["query"][0]["q_type"]
  flag = p["flags"]
  id   = p["id"]

  return [id.unpack("H*")[0], flag.unpack("H*")[0], type.unpack("H*")[0]].join(":")
end



###########################################################################

def security_check(packet_struct)
  method = {
    DnsAuthMD5    => "md5",
    DnsAuthSHA1   => "sha1",
    DnsAuthSHA256 => "sha256",
    DnsAuthSHA384 => "sha384",
    DnsAuthSHA512 => "sha512",
    DnsAuthRMD160 => "rmd160",
    DnsAuthPGP    => "pgp",
  }
  has_sec = 0
  auth_type = nil

  packet_struct["answer"].each {
    |i|
    name = dnsstr_to_str(i["a"]).downcase

    if ! (/^sec\-auth\-/ =~ name)
      if ENV["STRICT"] != nil && i["a_type"] == DnsType::A
	ML.log("STRICT: RR must be signed")
	return false
      end

      next
    end

    # CNAME ؤν̾ϺθˤʤΤǻߤ롣
    if i["a_type"] == DnsType::CNAME
      ML.log("CNAME sign can't be allowed")
      return false
    end

    # 䤤碌ɥᥤʳξ󤬴ޤޤƤƤϤʤ
    if ENV["STRICT"] != nil
      domain_makelist(name).each {
	|j|
	if ! (Regexp.new(Regexp.escape(j) + "$") =~ name ) #"
	  ML.log("STRICT: RR poison reject")
	  return false
	end
      }
    end

    auth = nil
    method.each {
      |j|
      if Regexp.new("^sec\-auth\-#{j[1]}\-\-") =~ name
	digest = j[0].new(i)
	if digest.valid() == false
	  return false
	end

	auth = j[1] # ǧˡĤä
	break
      end
    }
    if auth == nil
      ML.log("Unknown method")
      return false
    end

    # ǧˡƱΤ
    auth_type = auth if auth_type == nil
    if ENV["STRICT"] && auth_type != auth
      ML.log("Mixed auth")
      return false
    end

    has_sec += 1
  }

  if (has_sec == 0 && ENV["STRICT"] != nil)
    ML.log("STRICT: Must have one or more signed RR")
    return false
  end

  return true
end

#############################################################################
#  UDP Server
#############################################################################

# DNS Ф DNS Ф󶡤ɬפʲǽ TCP/UDPsocket
# 饹ȤʤΤǡ Socket 饹ľ롣
Udownlink = Socket.open(Socket::AF_INET, Socket::SOCK_DGRAM, 0)
Udownlink.bind([Socket::AF_INET, Service::Domain, str_to_ipstr(DOWNLINK), ""].pack("s n a4 a8"))
Uuplink   = Socket.open(Socket::AF_INET, Socket::SOCK_DGRAM, 0)

Session_cache = Hash.new

#------------------------------------------------------
# client => server
#------------------------------------------------------
Thread.start {
  loop {
    (client_packet, client_addr) = Udownlink.recvfrom(1024)

#ML.log("qstart")

    Thread.start {

      addr = client_addr[4,4]

      #p client_addr

      p = parse_packet(client_packet)
      if p == nil
	ML.log("packet parse failed. reject")
	Thread.exit
      end

      ML.log("q  " + addr.unpack("H*")[0] + ":" +
	     header_str(p) + " " +
	     dnsstr_to_str(p["query"][0]["q"]))

      Session_cache[make_key(p)] = {
	"addr"     => client_addr,
	"timeout" => Time.now.to_i + 60
      }

      Uuplink.send(client_packet, 0, [Socket::AF_INET, Service::Domain, str_to_ipstr(UPLINK), ""].pack("s n a4 a8"))

    }
  }
}


#------------------------------------------------------
# client <= server
#------------------------------------------------------
Thread.start {
  loop {
    (server_packet, server_addr) = Uuplink.recvfrom(1024)

#ML.log("astart")

    Thread.start {

      if ipstr_to_str(server_addr[4,4]) != UPLINK
	ML.log("server reply from wrong host: " +
		ipstr_to_str(server_addr[4,4]) + " / " + server_packet.dump)
	Thread.exit
      end

      p = parse_packet(server_packet)
      if p == nil
	ML.log("packet parse failed. reject")
	Thread.exit
      end

      key = make_key(p)
      if Session_cache[key] == nil
	ML.log("server reply not found in session cache: " + server_packet.dump)
	Thread.exit
      end

      client_addr = Session_cache[key]["addr"]

      # ǧڤ
      if p != nil && p["answer"] != nil
	r = security_check(p)

	if r == false
	  pac = make_reject_packet(p)
	  Udownlink.send(pac, 0, client_addr)

	  ML.log("Authentication failed. reject: " + server_packet.dump)

	  Session_cache.delete(key)
	  Thread.exit

	end
      end

      addr = client_addr[4,4]
      Udownlink.send(server_packet, 0, client_addr)

      ML.log("a  " + addr.unpack("H*")[0] + ":" +
	     header_str(p) + " " +
	     dnsstr_to_str(p["query"][0]["q"]) + " " +
	     num_records_str(p))

      if p["answer"] != nil
	p["answer"].each {
	  |i|
	  ML.log("rr " +
		 addr.unpack("H*")[0] + " " +
		 i["ttl"].unpack("N")[0].to_s + " " +
		 i["a_type"].unpack("n")[0].to_s + " " +
		 dnsstr_to_str(i["a"]) + " - " +
		 dnsdata_to_str(i["data"], i["a_type"])
		 )
	}
      end

      Session_cache.delete(key)

    }
  }
}


#############################################################################
# TCP server
#############################################################################

class DnsTcpAuthError < StandardError ; end

Thread.start {
  downlink_t = Socket.open(Socket::AF_INET, Socket::SOCK_STREAM, 0)
  downlink_t.bind([Socket::AF_INET, Service::Domain, str_to_ipstr(DOWNLINK), ""].pack("s n a4 a8"))
  downlink_t.listen(20) #  20 饤ȤޤƱ³

  loop {
    netsock_t = downlink_t.accept
    Thread.start {
      down_sock = netsock_t

      begin
	up_sock = nil
	timeout(60) {
	  up_sock = Socket.open(Socket::AF_INET, Socket::SOCK_STREAM, 0)
	  up_sock.connect([Socket::AF_INET, Service::Domain, str_to_ipstr(UPLINK), ""].pack("s n a4 a8"))

	  #------------------------------------------------------
	  # client => server
	  #------------------------------------------------------
	  client_len    = down_sock[0].read(2)
	  client_packet = down_sock[0].read(client_len.unpack("n")[0])

	  up_sock.write(client_len + client_packet)

	  client_addr = down_sock[1]
	  addr = client_addr[4,4]

	  p = parse_packet(client_packet)
	  ML.log("Q  " + addr.unpack("H*")[0] + ":" +
		 header_str(p) + " " +
		 dnsstr_to_str(p["query"][0]["q"]))


	  #------------------------------------------------------
	  # client <= server
	  #------------------------------------------------------
	  server_len    = up_sock.read(2)
	  server_packet = up_sock.read(server_len.unpack("n")[0])

	  #p server_packet
	  p = parse_packet(server_packet)

	  # ǧڤ
	  if p != nil && p["answer"] != nil
	    r = security_check(p)

	    if r == false
	      pac = make_reject_packet(p)
	      len = [pac.length].pack("n")

	      down_sock[0].write(len + pac)

	      ML.log("Authentication failed. reject: " + server_packet.dump)
	      raise DnsTcpAuthError
	    end
	  end

	  down_sock[0].write(server_len + server_packet)

	  ML.log("A  " + addr.unpack("H*")[0] + ":" +
		 header_str(p) + " " +
		 dnsstr_to_str(p["query"][0]["q"]) + " " +
		 num_records_str(p))

	  if p["answer"] != nil
	    p["answer"].each {
	      |i|
	      ML.log("RR " +
		     addr.unpack("H*")[0] + " " +
		     i["ttl"].unpack("N")[0].to_s + " " +
		     i["a_type"].unpack("n")[0].to_s + " " +
		     dnsstr_to_str(i["a"]) + " - " +
		     dnsdata_to_str(i["data"], i["a_type"])
		     )
	    }
	  end

	}
      rescue TimeoutError
	ML.log("TCP query timeout")
      rescue DnsTcpAuthError
	ML.log("TCP auth err")
      rescue
	ML.log("TCP other err")
      ensure
	up_sock.close
	down_sock[0].close
      end
    }
  }
}

#########################################################################

# å
loop {
  sleep 10
  cache_timeout(Session_cache, "session timeout: ")

  # å˥¤򤫤
  sleep 10
  cache_reduction(Session_cache, 500, "session overflow: ")
}

# end
